from syrtis.shell import *
from syrtis.material import *
from syrtis.habitat import *
from syrtis.configuration import *
from syrtis.solver import *
from syrtis.solvermanager import *
from syrtis.tools import *

import syrtis.shell
import syrtis.material
import syrtis.habitat